@@ -93,15 +93,18 @@ struct eventcomm_proto_data {
     int have_monotonic_clock;
 };
 
+#ifdef HAVE_LIBEVDEV_DEVICE_LOG_FUNCS
 static void
-libevdev_log_func(enum libevdev_log_priority priority,
+libevdev_log_func(const struct libevdev *dev,
+                  enum libevdev_log_priority priority,
                   void *data,
                   const char *file, int line, const char *func,
                   const char *format, va_list args)
-_X_ATTRIBUTE_PRINTF(6, 0);
+_X_ATTRIBUTE_PRINTF(7, 0);
 
 static void
-libevdev_log_func(enum libevdev_log_priority priority,
+libevdev_log_func(const struct libevdev *dev,
+                  enum libevdev_log_priority priority,
                   void *data,
                   const char *file, int line, const char *func,
                   const char *format, va_list args)
@@ -111,19 +114,15 @@ libevdev_log_func(enum libevdev_log_priority priority,
     switch(priority) {
         case LIBEVDEV_LOG_ERROR: verbosity = 0; break;
         case LIBEVDEV_LOG_INFO: verbosity = 4; break;
-        case LIBEVDEV_LOG_DEBUG: verbosity = 10; break;
+        case LIBEVDEV_LOG_DEBUG:
+        default:
+            verbosity = 10;
+            break;
     }
 
     LogVMessageVerbSigSafe(X_NOTICE, verbosity, format, args);
 }
-
-static void
-set_libevdev_log_handler(void)
-{
-                              /* be quiet, gcc *handwave* */
-    libevdev_set_log_function((libevdev_log_func_t)libevdev_log_func, NULL);
-    libevdev_set_log_priority(LIBEVDEV_LOG_DEBUG);
-}
+#endif
 
 struct eventcomm_proto_data *
 EventProtoDataAlloc(int fd)
@@ -131,7 +130,6 @@ EventProtoDataAlloc(int fd)
     struct eventcomm_proto_data *proto_data;
     int rc;
 
-    set_libevdev_log_handler();
 
     proto_data = calloc(1, sizeof(struct eventcomm_proto_data));
     if (!proto_data)
@@ -140,12 +138,31 @@ EventProtoDataAlloc(int fd)
     proto_data->st_to_mt_scale[0] = 1;
     proto_data->st_to_mt_scale[1] = 1;
 
-    rc = libevdev_new_from_fd(fd, &proto_data->evdev);
+    proto_data->evdev = libevdev_new();
+    if (!proto_data->evdev) {
+        rc = -1;
+        goto out;
+    }
+
+#ifdef HAVE_LIBEVDEV_DEVICE_LOG_FUNCS
+    libevdev_set_device_log_function(proto_data->evdev, libevdev_log_func,
+                                     LIBEVDEV_LOG_DEBUG, NULL);
+#endif
+
+    rc = libevdev_set_fd(proto_data->evdev, fd);
     if (rc < 0) {
+        goto out;
+    }
+
+    proto_data->read_flag = LIBEVDEV_READ_FLAG_NORMAL;
+
+out:
+    if (rc < 0) {
+        if (proto_data && proto_data->evdev)
+            libevdev_free(proto_data->evdev);
         free(proto_data);
         proto_data = NULL;
-    } else
-        proto_data->read_flag = LIBEVDEV_READ_FLAG_NORMAL;
+    }
 
     return proto_data;
 }
@@ -222,8 +239,6 @@ EventDeviceOnHook(InputInfoPtr pInfo, SynapticsParameters * para)
         (struct eventcomm_proto_data *) priv->proto_data;
     int ret;
 
-    set_libevdev_log_handler();
-
     if (libevdev_get_fd(proto_data->evdev) != -1) {
         struct input_event ev;
 
@@ -284,58 +299,42 @@ EventDeviceOffHook(InputInfoPtr pInfo)
  * - BTN_TOOL_PEN is _not_ set
  *
  * @param evdev Libevdev handle
- * @param test_grab If true, test whether an EVIOCGRAB is possible on the
- * device. A failure to grab the event device returns in a failure.
  *
  * @return TRUE if the device is a touchpad or FALSE otherwise.
  */
 static Bool
-event_query_is_touchpad(struct libevdev *evdev, BOOL test_grab)
+event_query_is_touchpad(struct libevdev *evdev)
 {
-    int ret = FALSE, rc;
-
-    if (test_grab) {
-        rc = libevdev_grab(evdev, LIBEVDEV_GRAB);
-        if (rc < 0)
-            return FALSE;
-    }
-
     /* Check for ABS_X, ABS_Y, ABS_PRESSURE and BTN_TOOL_FINGER */
     if (!libevdev_has_event_type(evdev, EV_SYN) ||
         !libevdev_has_event_type(evdev, EV_ABS) ||
         !libevdev_has_event_type(evdev, EV_KEY))
-        goto unwind;
+        return FALSE;
 
     if (!libevdev_has_event_code(evdev, EV_ABS, ABS_X) ||
         !libevdev_has_event_code(evdev, EV_ABS, ABS_Y))
-        goto unwind;
+        return FALSE;
 
     /* we expect touchpad either report raw pressure or touches */
     if (!libevdev_has_event_code(evdev, EV_KEY, BTN_TOUCH) &&
         !libevdev_has_event_code(evdev, EV_ABS, ABS_PRESSURE))
-        goto unwind;
+        return FALSE;
 
     /* all Synaptics-like touchpad report BTN_TOOL_FINGER */
     if (!libevdev_has_event_code(evdev, EV_KEY, BTN_TOOL_FINGER) ||
         libevdev_has_event_code(evdev, EV_ABS, BTN_TOOL_PEN)) /* Don't match wacom tablets */
-        goto unwind;
+        return FALSE;
 
     if (libevdev_has_event_code(evdev, EV_ABS, ABS_MT_SLOT)) {
         if (libevdev_get_num_slots(evdev) == -1)
-            goto unwind; /* Ignore fake MT devices */
+            return FALSE; /* Ignore fake MT devices */
 
         if (!libevdev_has_event_code(evdev, EV_ABS, ABS_MT_POSITION_X) ||
             !libevdev_has_event_code(evdev, EV_ABS, ABS_MT_POSITION_Y))
-            goto unwind;
+            return FALSE;
     }
 
-    ret = TRUE;
-
- unwind:
-    if (test_grab)
-        libevdev_grab(evdev, LIBEVDEV_UNGRAB);
-
-    return (ret == TRUE);
+    return TRUE;
 }
 
 #define PRODUCT_ANY 0x0000
@@ -537,8 +536,7 @@ EventQueryHardware(InputInfoPtr pInfo)
     SynapticsPrivate *priv = (SynapticsPrivate *) pInfo->private;
     struct eventcomm_proto_data *proto_data = priv->proto_data;
 
-    if (!event_query_is_touchpad(proto_data->evdev,
-                                 (proto_data) ? proto_data->need_grab : TRUE))
+    if (!event_query_is_touchpad(proto_data->evdev))
         return FALSE;
 
     xf86IDrvMsg(pInfo, X_PROBED, "touchpad found\n");
@@ -559,6 +557,7 @@ SynapticsReadEvent(InputInfoPtr pInfo, struct input_event *ev)
         if (rc != -EAGAIN) {
             LogMessageVerbSigSafe(X_ERROR, 0, "%s: Read error %d\n", pInfo->name,
                     errno);
+            xf86RemoveEnabledDevice(pInfo);
         } else if (proto_data->read_flag == LIBEVDEV_READ_FLAG_SYNC) {
             proto_data->read_flag = LIBEVDEV_READ_FLAG_NORMAL;
             return SynapticsReadEvent(pInfo, ev);
@@ -627,6 +626,14 @@ EventProcessTouchEvent(InputInfoPtr pInfo, struct SynapticsHwState *hw,
                 hw->slot_state[slot_index] = SLOTSTATE_CLOSE;
                 proto_data->num_touches--;
             }
+
+            /* When there are no fingers on the touchpad, set width and
+             * pressure to zero as ABS_MT_TOUCH_MAJOR and ABS_MT_PRESSURE
+             * are not zero when fingers are released. */
+            if (proto_data->num_touches == 0) {
+                hw->fingerWidth = 0;
+                hw->z = 0;
+            }
         }
         else {
             ValuatorMask *mask = proto_data->last_mt_vals[slot_index];
@@ -639,6 +646,12 @@ EventProcessTouchEvent(InputInfoPtr pInfo, struct SynapticsHwState *hw,
                     hw->cumulative_dx += ev->value - last_val;
                 else if (ev->code == ABS_MT_POSITION_Y)
                     hw->cumulative_dy += ev->value - last_val;
+                else if (ev->code == ABS_MT_TOUCH_MAJOR &&
+                         priv->has_mt_palm_detect)
+                    hw->fingerWidth = ev->value;
+                else if (ev->code == ABS_MT_PRESSURE &&
+                         priv->has_mt_palm_detect)
+                    hw->z = ev->value;
             }
 
             valuator_mask_set(mask, map, ev->value);
@@ -693,14 +706,12 @@ EventReadHwState(InputInfoPtr pInfo,
     struct eventcomm_proto_data *proto_data = priv->proto_data;
     Bool sync_cumulative = FALSE;
 
-    set_libevdev_log_handler();
-
     SynapticsResetTouchHwState(hw, FALSE);
 
     /* Reset cumulative values if buttons were not previously pressed and no
      * two-finger scrolling is ongoing, or no finger was previously present. */
     if (((!hw->left && !hw->right && !hw->middle) &&
-        !(priv->vert_scroll_twofinger_on || priv->vert_scroll_twofinger_on)) ||
+        !(priv->vert_scroll_twofinger_on || priv->horiz_scroll_twofinger_on)) ||
         hw->z < para->finger_low) {
         hw->cumulative_dx = hw->x;
         hw->cumulative_dy = hw->y;
@@ -722,6 +733,10 @@ EventReadHwState(InputInfoPtr pInfo,
             }
             break;
         case EV_KEY:
+            /* ignore hw repeat events */
+            if (ev.value > 1)
+                break;
+
             v = (ev.value ? TRUE : FALSE);
             switch (ev.code) {
             case BTN_LEFT:
@@ -891,6 +906,10 @@ event_query_touch(InputInfoPtr pInfo)
             return;
         }
 
+        if (libevdev_has_event_code(dev, EV_ABS, ABS_MT_TOUCH_MAJOR) &&
+            libevdev_has_event_code(dev, EV_ABS, ABS_MT_PRESSURE))
+            priv->has_mt_palm_detect = TRUE;
+
         axnum = 0;
         for (axis = ABS_MT_SLOT + 1; axis <= ABS_MT_MAX; axis++) {
             int axis_idx = axis - ABS_MT_TOUCH_MAJOR;
@@ -951,7 +970,7 @@ EventReadDevDimensions(InputInfoPtr pInfo)
         proto_data->axis_map[i] = -1;
     proto_data->cur_slot = -1;
 
-    if (event_query_is_touchpad(proto_data->evdev, proto_data->need_grab)) {
+    if (event_query_is_touchpad(proto_data->evdev)) {
         event_query_touch(pInfo);
         event_query_axis_ranges(pInfo);
     }
@@ -985,19 +1004,19 @@ EventAutoDevProbe(InputInfoPtr pInfo, const char *device)
 
             rc = libevdev_new_from_fd(fd, &evdev);
             if (rc >= 0) {
-                touchpad_found = event_query_is_touchpad(evdev, TRUE);
+                touchpad_found = event_query_is_touchpad(evdev);
                 libevdev_free(evdev);
             }
 
             if (!(pInfo->flags & XI86_SERVER_FD))
                 SYSCALL(close(fd));
-
-            /* if a device is set and not a touchpad (or already grabbed),
-             * we must return FALSE.  Otherwise, we'll add a device that
-             * wasn't requested for and repeat
-             * f5687a6741a19ef3081e7fd83ac55f6df8bcd5c2. */
-            return touchpad_found;
         }
+
+        /* if a device is set and not a touchpad (or already grabbed),
+         * we must return FALSE.  Otherwise, we'll add a device that
+         * wasn't requested for and repeat
+         * f5687a6741a19ef3081e7fd83ac55f6df8bcd5c2. */
+        return touchpad_found;
     }
 
     i = scandir(DEV_INPUT_EVENT, &namelist, EventDevOnly, alphasort);
@@ -1027,7 +1046,7 @@ EventAutoDevProbe(InputInfoPtr pInfo, const char *device)
 
             rc = libevdev_new_from_fd(fd, &evdev);
             if (rc >= 0) {
-                touchpad_found = event_query_is_touchpad(evdev, TRUE);
+                touchpad_found = event_query_is_touchpad(evdev);
                 libevdev_free(evdev);
                 if (touchpad_found) {
                     xf86IDrvMsg(pInfo, X_PROBED, "auto-dev sets device to %s\n",
